/*************************************************************************************************
 * Utility interface
 *
 * Copyright 2020 Google LLC
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file
 * except in compliance with the License.  You may obtain a copy of the License at
 *     https://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software distributed under the
 * License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
 * either express or implied.  See the License for the specific language governing permissions
 * and limitations under the License.
 *************************************************************************************************/

package tkrzw_rpc

import (
	"fmt"
	"reflect"
	"strconv"
	"strings"
	"unsafe"
)

// The special bytes value for no-operation or any data.
var AnyBytes = []byte("\x00[ANY]\x00")

// The special string value for no-operation or any data.
var AnyString = string([]byte("\x00[ANY]\x00"))

// The special string value for non-existing data.
var NilString = string([]byte("\x00[NIL]\x00"))

// Converts any object into a string.
//
// @param x The object to convert.
// @return The result string.
func ToString(x interface{}) string {
	switch x := x.(type) {
	case []byte:
		return string(x)
	case string:
		return x
	case bool:
		if x {
			return "true"
		}
		return "false"
	case int, uint, int8, uint8, int16, uint16, int32, uint32, int64, uint64:
		return fmt.Sprintf("%d", x)
	case uintptr:
		return fmt.Sprintf("0x%016x", x)
	case float32, float64, complex64, complex128:
		return fmt.Sprintf("%.6f", x)
	case error:
		return x.Error()
	}
	if x == nil {
		return ""
	}
	if str, ok := x.(fmt.Stringer); ok {
		return str.String()
	}
	return fmt.Sprintf("#<%T>", x)
}

// Converts any object into a byte array.
//
// @param x The object to convert.
// @return The result byte array.
func ToByteArray(x interface{}) []byte {
	switch x := x.(type) {
	case []byte:
		return x
	case string:
		return []byte(x)
	case bool:
		if x {
			return []byte("true")
		}
		return []byte("false")
	case int, uint, int8, uint8, int16, uint16, int32, uint32, int64, uint64:
		return []byte(fmt.Sprintf("%d", x))
	case uintptr:
		return []byte(fmt.Sprintf("0x%016x", x))
	case float32, float64, complex64, complex128:
		return []byte(fmt.Sprintf("%.6f", x))
	case error:
		return []byte(x.Error())
	}
	if x == nil {
		return make([]byte, 0)
	}
	if str, ok := x.(fmt.Stringer); ok {
		return []byte(str.String())
	}
	return []byte(fmt.Sprintf("#<%T>", x))
}

// Converts any object into an integer.
//
// @param x The object to convert.
// @return The result integer.
func ToInt(value interface{}) int64 {
	switch value := value.(type) {
	case int:
		return int64(value)
	case uint:
		return int64(value)
	case int8:
		return int64(value)
	case uint8:
		return int64(value)
	case int16:
		return int64(value)
	case uint16:
		return int64(value)
	case int32:
		return int64(value)
	case uint32:
		return int64(value)
	case int64:
		return int64(value)
	case uint64:
		return int64(value)
	case float32:
		return int64(value)
	case float64:
		return int64(value)
	case string:
		int_value, err := strconv.ParseInt(value, 10, 64)
		if err == nil {
			return int64(int_value)
		}
		float_value, err := strconv.ParseFloat(value, 64)
		if err == nil {
			return int64(float_value)
		}
	case []byte:
		int_value, err := strconv.ParseInt(string(value), 10, 64)
		if err == nil {
			return int64(int_value)
		}
		float_value, err := strconv.ParseFloat(string(value), 64)
		if err == nil {
			return int64(float_value)
		}
	case bool:
		if value {
			return 1
		}
	}
	return 0
}

// Converts any object into a real number.
//
// @param x The object to convert.
// @return The result real number.
func ToFloat(value interface{}) float64 {
	switch value := value.(type) {
	case int:
		return float64(value)
	case uint:
		return float64(value)
	case int8:
		return float64(value)
	case uint8:
		return float64(value)
	case int16:
		return float64(value)
	case uint16:
		return float64(value)
	case int32:
		return float64(value)
	case uint32:
		return float64(value)
	case int64:
		return float64(value)
	case uint64:
		return float64(value)
	case float32:
		return float64(value)
	case float64:
		return float64(value)
	case string:
		int_value, err := strconv.ParseFloat(value, 64)
		if err == nil {
			return float64(int_value)
		}
	case []byte:
		int_value, err := strconv.ParseFloat(string(value), 64)
		if err == nil {
			return float64(int_value)
		}
	case bool:
		if value {
			return 1.0
		}
	}
	return 0.0
}

// Checks whether the given data is a unique value of any data.
//
// @param data The data to check.
// @return True if the data is any data or, false if not.
func IsAnyData(data interface{}) bool {
	switch data := data.(type) {
	case []byte:
		return IsAnyBytes(data)
	case string:
		return IsAnyString(data)
	default:
		return false
	}
}

// Checks whether the given bytes are the any bytes.
//
// @param data The data to check.
// @return True if the data are the any bytes.
func IsAnyBytes(data []byte) bool {
	return ((*reflect.SliceHeader)(unsafe.Pointer(&data)).Data ==
		(*reflect.SliceHeader)(unsafe.Pointer(&AnyBytes)).Data)
}

// Checks whether the given string is the any string.
//
// @param data The data to check.
// @return True if the data is the string, or false if not.
func IsAnyString(data string) bool {
	return ((*reflect.StringHeader)(unsafe.Pointer(&data)).Data ==
		(*reflect.StringHeader)(unsafe.Pointer(&AnyString)).Data)
}

// Checks whether the given data is a nil-equivalent value.
//
// @param data The data to check.
// @return True if the data is a nil-equivalent value, or false if not.
func IsNilData(data interface{}) bool {
	switch data := data.(type) {
	case string:
		return IsNilString(data)
	}
	return data == nil
}

// Checks whether the given string is the nil string.
//
// @param data The data to check.
// @return True if the data is the nil string, or false if not.
func IsNilString(data string) bool {
	return ((*reflect.StringHeader)(unsafe.Pointer(&data)).Data ==
		(*reflect.StringHeader)(unsafe.Pointer(&NilString)).Data)
}

// Parses a parameter string to make a parameter string map.
//
// @param expr A parameter string in "name=value,name=value,..." format.
// @return The string map of the parameters.
func ParseParams(expr string) map[string]string {
	params := make(map[string]string)
	fields := strings.Split(expr, ",")
	for _, field := range fields {
		columns := strings.SplitN(field, "=", 2)
		if len(columns) != 2 {
			continue
		}
		name := strings.TrimSpace(columns[0])
		value := strings.TrimSpace(columns[1])
		if len(name) > 0 {
			params[name] = value
		}
	}
	return params
}

// END OF FILE
